@@ -1,7 +1,7 @@
 # datasets.py
 
 from geode.utilities import convert_labels_to_one_hots, rasterize_polygon_layer, resample_dataset, tile_raster_pair
-from numpy import abs, asarray, flip, float32, moveaxis, rot90
+from numpy import abs, flip, moveaxis, rot90
 from numpy.testing import assert_allclose
 from numpy.random import randint, shuffle
 from os import listdir, mkdir
@@ -10,7 +10,7 @@ from osgeo import gdal, ogr
 import tensorflow as tf
 
 
-class Segmentation:
+class SegmentationDataset:
     """Defines a semantic segmentation dataset to be used in deep-learning models. Has methods to resample source
      imagery, to rasterize polygon layers, to generate training tiles, and to generate an iterator object
      for model training."""
