@@ -7,10 +7,10 @@ from os import listdir, makedirs
 from os.path import isdir, join
 from osgeo.gdal import Open
 import tensorflow as tf
-from tensorflow.keras.layers import BatchNormalization, Concatenate, Conv2D, Dropout, MaxPooling2D, UpSampling2D
+from tensorflow.keras.layers import BatchNormalization, Concatenate, Conv2D, Dropout, Input, MaxPooling2D, UpSampling2D
 
 
-class SegmentationModel(tf.keras.Model):
+class SegmentationModel:
 
     def __init__(self):
 
@@ -18,6 +18,7 @@ class SegmentationModel(tf.keras.Model):
 
         self.test_metrics = {}
         self.test_filenames = []
+        self.model = None
 
     def compute_metrics(self, test_labels_path: str = None,
                         test_predictions_path: str = None,
@@ -133,7 +134,7 @@ class SegmentationModel(tf.keras.Model):
             rgb = Open(join(test_imagery_path, fname))
 
             predict_raster(input_dataset=rgb,
-                           model=self,
+                           model=self.model,
                            output_path=join(test_predictions_path, fname))
 
             # close the input dataset
@@ -149,9 +150,9 @@ class Unet(SegmentationModel):
     def __init__(self, n_channels: int = 3,
                  n_classes: int = 2,
                  n_filters: int = 16,
-                 dropout_rate: float = 0.2):
+                 dropout_rate: float = 0.3):
 
-        # initialize the Model superclass
+        # initialize the superclass
         super().__init__()
 
         # define attributes
@@ -160,33 +161,35 @@ class Unet(SegmentationModel):
         self.n_filters = n_filters
         self.dropout_rate = dropout_rate
 
-        # Multiple layer versions are required because they get called on different input shapes
+        # define all the Unet layers
 
         # Downsampling-path convolutional layers
+        self.inputs = Input(shape=(None, None, n_channels))
+
         self.conv_down_0 = [Conv2D(filters=self.n_filters,
-                            kernel_size=(3, 3),
-                            padding='same',
-                            activation='relu') for i in range(2)]
+                                   kernel_size=(3, 3),
+                                   padding='same',
+                                   activation='relu') for i in range(2)]
 
         self.conv_down_1 = [Conv2D(filters=2 * self.n_filters,
-                            kernel_size=(3, 3),
-                            padding='same',
-                            activation='relu') for i in range(2)]
+                                   kernel_size=(3, 3),
+                                   padding='same',
+                                   activation='relu') for i in range(2)]
 
         self.conv_down_2 = [Conv2D(filters=4 * self.n_filters,
-                            kernel_size=(3, 3),
-                            padding='same',
-                            activation='relu') for i in range(4)]
+                                   kernel_size=(3, 3),
+                                   padding='same',
+                                   activation='relu') for i in range(4)]
 
         self.conv_down_3 = [Conv2D(filters=8 * self.n_filters,
-                            kernel_size=(3, 3),
-                            padding='same',
-                            activation='relu') for i in range(4)]
+                                   kernel_size=(3, 3),
+                                   padding='same',
+                                   activation='relu') for i in range(4)]
 
         self.conv_down_4 = [Conv2D(filters=8 * self.n_filters,
-                            kernel_size=(3, 3),
-                            padding='same',
-                            activation='relu') for i in range(4)]
+                                   kernel_size=(3, 3),
+                                   padding='same',
+                                   activation='relu') for i in range(4)]
 
         self.conv_up_3 = [Conv2D(filters=8 * self.n_filters,
                                  kernel_size=(3, 3),
@@ -222,10 +225,10 @@ class Unet(SegmentationModel):
 
         # Max-pooling layers
         self.max_pooling = [MaxPooling2D(pool_size=(2, 2),
-                                         padding='same') for i in range(5)]
+                                         padding='same') for i in range(4)]
 
         # Upsampling layers
-        self.upsampling = [UpSampling2D(size=(2, 2)) for i in range(5)]
+        self.upsampling = [UpSampling2D(size=(2, 2)) for i in range(4)]
 
         # Batch normalization layers
         self.batch_normalization = [BatchNormalization() for i in range(n_do_bn)]
@@ -234,94 +237,139 @@ class Unet(SegmentationModel):
         self.dropout = [Dropout(rate=self.dropout_rate) for i in range(n_do_bn)]
 
         # Concatenate layers
-        self.concatenate = [Concatenate(axis=-1) for i in range(5)]
+        self.concatenate = [Concatenate(axis=-1) for i in range(4)]
 
-    def call(self, input_tensor,
-             training=True):
+    def compile_model(self, loss: tf.keras.losses.Loss = 'sparse_categorical_crossentropy',
+                      learning_rate: float = 0.0001) -> None:
 
-        include_dropout = training and self.dropout_rate == 0.0
-        conv_counter = 0
+        """Returns a model object, compiled with the provided loss and optimizer. Additionally, this sets the self.model
+        attribute with the compiled model.
 
-        # downsampling path
+        Args:
+            loss: the loss function to use during training;
+            learning_rate: the starting learning rate for the Adam optimizer.
+
+        Returns:
+            None"""
+
+        include_dropout = (self.dropout_rate > 0.0)
+
+        # build the model graph
 
         # level 0
-        d0 = input_tensor
-        for i in range(2):
-            d0 = self.conv_down_0[i](d0)
-            d0 = self.dropout[conv_counter](d0) if include_dropout else d0
-            d0 = self.batch_normalization[conv_counter](d0)
-            conv_counter += 1
+        inputs = Input(shape=(None, None, self.n_channels), dtype=tf.float32)
+        d0 = Conv2D(filters=self.n_filters,
+                    kernel_size=(3, 3),
+                    padding='same',
+                    activation='relu')(inputs)
+        d0 = Dropout(rate=self.dropout_rate)(d0) if include_dropout else d0
+        d0 = BatchNormalization()(d0)
+        d0 = Conv2D(filters=self.n_filters,
+                    kernel_size=(3, 3),
+                    padding='same',
+                    activation='relu')(d0)
+        d0 = Dropout(rate=self.dropout_rate)(d0) if include_dropout else d0
+        d0 = BatchNormalization()(d0)
 
         # level 1
-        d1 = self.max_pooling[0](d0)
+        d1 = MaxPooling2D(pool_size=(2, 2),
+                          padding='same')(d0)
         for i in range(2):
-            d1 = self.conv_down_1[i](d1)
-            d1 = self.dropout[conv_counter](d1) if include_dropout else d1
-            d1 = self.batch_normalization[conv_counter](d1)
-            conv_counter += 1
+            d1 = Conv2D(filters=2 * self.n_filters,
+                        kernel_size=(3, 3),
+                        padding='same',
+                        activation='relu')(d1)
+            d1 = Dropout(rate=self.dropout_rate)(d1) if include_dropout else d1
+            d1 = BatchNormalization()(d1)
 
         # level 2
-        d2 = self.max_pooling[1](d1)
+        d2 = MaxPooling2D(pool_size=(2, 2),
+                          padding='same')(d1)
         for i in range(4):
-            d2 = self.conv_down_2[i](d2)
-            d2 = self.dropout[conv_counter](d2) if include_dropout else d2
-            d2 = self.batch_normalization[conv_counter](d2)
-            conv_counter += 1
+            d2 = Conv2D(filters=4 * self.n_filters,
+                        kernel_size=(3, 3),
+                        padding='same',
+                        activation='relu')(d2)
+            d2 = Dropout(rate=self.dropout_rate)(d2) if include_dropout else d2
+            d2 = BatchNormalization()(d2)
 
         # level 3
-        d3 = self.max_pooling[2](d2)
+        d3 = MaxPooling2D(pool_size=(2, 2),
+                          padding='same')(d2)
         for i in range(4):
-            d3 = self.conv_down_3[i](d3)
-            d3 = self.dropout[conv_counter](d3) if include_dropout else d3
-            d3 = self.batch_normalization[conv_counter](d3)
-            conv_counter += 1
+            d3 = Conv2D(filters=8 * self.n_filters,
+                        kernel_size=(3, 3),
+                        padding='same',
+                        activation='relu')(d3)
+            d3 = Dropout(rate=self.dropout_rate)(d3) if include_dropout else d3
+            d3 = BatchNormalization()(d3)
 
         # level 4
-        d4 = self.max_pooling[3](d3)
+        d4 = MaxPooling2D(pool_size=(2, 2),
+                          padding='same')(d3)
         for i in range(4):
-            d4 = self.conv_down_4[i](d4)
-            d4 = self.dropout[conv_counter](d4) if include_dropout else d4
-            d4 = self.batch_normalization[conv_counter](d4)
-            conv_counter += 1
+            d4 = Conv2D(filters=8 * self.n_filters,
+                        kernel_size=(3, 3),
+                        padding='same',
+                        activation='relu')(d4)
+            d4 = Dropout(rate=self.dropout_rate)(d4) if include_dropout else d4
+            d4 = BatchNormalization()(d4)
 
         # upsampling path
 
         # level 3
-        u3 = self.upsampling[3](d4)
-        u3 = self.concatenate[3]([u3, d3])
+        u3 = UpSampling2D(size=(2, 2))(d4)
+        u3 = Concatenate(axis=-1)([u3, d3])
         for i in range(4):
-            u3 = self.conv_up_3[i](u3)
-            u3 = self.dropout[conv_counter](u3) if include_dropout else u3
-            u3 = self.batch_normalization[conv_counter](u3)
-            conv_counter += 1
+            u3 = Conv2D(filters=8 * self.n_filters,
+                        kernel_size=(3, 3),
+                        padding='same',
+                        activation='relu')(u3)
+            u3 = Dropout(rate=self.dropout_rate)(u3) if include_dropout else u3
+            u3 = BatchNormalization()(u3)
 
         # level 2
-        u2 = self.upsampling[2](u3)
-        u2 = self.concatenate[2]([u2, d2])
+        u2 = UpSampling2D(size=(2, 2))(u3)
+        u2 = Concatenate(axis=-1)([u2, d2])
         for i in range(4):
-            u2 = self.conv_up_2[i](u2)
-            u2 = self.dropout[conv_counter](u2) if include_dropout else u2
-            u2 = self.batch_normalization[conv_counter](u2)
-            conv_counter += 1
+            u2 = Conv2D(filters=4 * self.n_filters,
+                        kernel_size=(3, 3),
+                        padding='same',
+                        activation='relu')(u2)
+            u2 = Dropout(rate=self.dropout_rate)(u2) if include_dropout else u2
+            u2 = BatchNormalization()(u2)
 
         # level 1
-        u1 = self.upsampling[1](u2)
-        u1 = self.concatenate[1]([u1, d1])
+        u1 = UpSampling2D(size=(2, 2))(u2)
+        u1 = Concatenate(axis=-1)([u1, d1])
         for i in range(2):
-            u1 = self.conv_up_1[i](u1)
-            u1 = self.dropout[conv_counter](u1) if include_dropout else u1
-            u1 = self.batch_normalization[conv_counter](u1)
-            conv_counter += 1
+            u1 = Conv2D(filters=2 * self.n_filters,
+                        kernel_size=(3, 3),
+                        padding='same',
+                        activation='relu')(u1)
+            u1 = Dropout(rate=self.dropout_rate)(u1) if include_dropout else u1
+            u1 = BatchNormalization()(u1)
 
         # level 0
-        u0 = self.upsampling[0](u1)
-        u0 = self.concatenate[0]([u0, d0])
+        u0 = UpSampling2D(size=(2, 2))(u1)
+        u0 = Concatenate(axis=-1)([u0, d0])
         for i in range(2):
-            u0 = self.conv_up_0[i](u0)
-            u0 = self.dropout[conv_counter](u0) if include_dropout else u0
-            u0 = self.batch_normalization[conv_counter](u0)
-            conv_counter += 1
+            u0 = Conv2D(filters=self.n_filters,
+                        kernel_size=(3, 3),
+                        padding='same',
+                        activation='relu')(u0)
+            u0 = Dropout(rate=self.dropout_rate)(u0) if include_dropout else u0
+            u0 = BatchNormalization()(u0)
 
-        output = self.conv_final(u0)
+        outputs = Conv2D(filters=self.n_classes,
+                         kernel_size=(1, 1),
+                         padding='same',
+                         activation='softmax')(u0)
 
-        return output
+        # create the model object
+        model = tf.keras.Model(inputs=inputs, outputs=outputs)
+
+        # compile the model
+        model.compile(loss=loss, optimizer=tf.keras.optimizers.Adam(learning_rate=learning_rate))
+
+        self.model = model
